
/* Chrysalide - Outil d'analyse de fichiers binaires
 * arithmop.c - gestion des opérations arithmétiques
 *
 * Copyright (C) 2022 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "arithmop.h"


#include "arithmop-int.h"
#include "literal.h"



/* --------------------- INTRODUCTION D'UNE NOUVELLE EXPRESSION --------------------- */


/* Initialise la classe des opérations arithmétiques. */
static void g_arithmetic_operation_class_init(GArithmOperationClass *);

/* Initialise une instance d'opération arithmétique. */
static void g_arithmetic_operation_init(GArithmOperation *);

/* Supprime toutes les références externes. */
static void g_arithmetic_operation_dispose(GArithmOperation *);

/* Procède à la libération totale de la mémoire. */
static void g_arithmetic_operation_finalize(GArithmOperation *);



/* --------------------- IMPLEMENTATION DES FONCTIONS DE CLASSE --------------------- */


/* Réalise une comparaison entre objets selon un critère précis. */
static bool g_arithmetic_operation_compare_rich(const GArithmOperation *, const GArithmOperation *, RichCmpOperation, bool *);

/* Reproduit une expression en place dans une nouvelle instance. */
static void g_arithmetic_operation_copy(GArithmOperation *, const GArithmOperation *);

/* Réduit une expression à une forme plus simple. */
static bool g_arithmetic_operation_reduce(GArithmOperation *, GScanContext *, GScanScope *, GScanExpression **);



/* ---------------------------------------------------------------------------------- */
/*                       INTRODUCTION D'UNE NOUVELLE EXPRESSION                       */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour une opération de relation entre expressions. */
G_DEFINE_TYPE(GArithmOperation, g_arithmetic_operation, G_TYPE_SCAN_EXPRESSION);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des opérations arithmétiques.           *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arithmetic_operation_class_init(GArithmOperationClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GScanExpressionClass *expr;             /* Version de classe parente   */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_arithmetic_operation_dispose;
    object->finalize = (GObjectFinalizeFunc)g_arithmetic_operation_finalize;

    expr = G_SCAN_EXPRESSION_CLASS(klass);

    expr->cmp_rich = (compare_expr_rich_fc)g_arithmetic_operation_compare_rich;
    expr->copy = (copy_expr_fc)g_arithmetic_operation_copy;
    expr->reduce = (reduce_expr_fc)g_arithmetic_operation_reduce;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : op = instance à initialiser.                                 *
*                                                                             *
*  Description : Initialise une instance d'opération arithmétique.            *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arithmetic_operation_init(GArithmOperation *op)
{
    op->left = NULL;
    op->right = NULL;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : op = instance d'objet GLib à traiter.                        *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arithmetic_operation_dispose(GArithmOperation *op)
{
    g_clear_object(&op->left);
    g_clear_object(&op->right);

    G_OBJECT_CLASS(g_arithmetic_operation_parent_class)->dispose(G_OBJECT(op));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : op = instance d'objet GLib à traiter.                        *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arithmetic_operation_finalize(GArithmOperation *op)
{
    G_OBJECT_CLASS(g_arithmetic_operation_parent_class)->finalize(G_OBJECT(op));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operator = type d'opération arithmétique à représenter.      *
*                left     = premier opérande concerné.                        *
*                right    = éventuel second opérande impliqué ou NULL.        *
*                                                                             *
*  Description : Organise une opération arithmétique entre expressions.       *
*                                                                             *
*  Retour      : Fonction mise en place.                                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GScanExpression *g_arithmetic_operation_new(ArithmeticExpressionOperator operator, GScanExpression *left, GScanExpression *right)
{
    GScanExpression *result;                /* Structure à retourner       */

    result = g_object_new(G_TYPE_ARITHMETIC_OPERATION, NULL);

    if (!g_arithmetic_operation_create(G_ARITHMETIC_OPERATION(result), operator, left, right))
        g_clear_object(&result);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : op       = instance à initialiser pleinement.                *
*                operator = type d'opération booléenne à représenter.         *
*                left     = premier opérande concerné.                        *
*                right    = éventuel second opérande impliqué ou NULL.        *
*                                                                             *
*  Description : Met en place une opération arithmétique entre expressions.   *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_arithmetic_operation_create(GArithmOperation *op, ArithmeticExpressionOperator operator, GScanExpression *left, GScanExpression *right)
{
    bool result;                            /* Bilan à retourner           */
    ExprValueType vtype;                    /* Type de valeur portée       */

    result = false;

    vtype = g_scan_expression_get_value_type(left);

    if (vtype != EVT_INTEGER && vtype != EVT_PENDING)
        goto exit;

    vtype = g_scan_expression_get_value_type(right);

    if (vtype != EVT_INTEGER && vtype != EVT_PENDING)
        goto exit;

    if (!g_scan_expression_create(G_SCAN_EXPRESSION(op), EVT_INTEGER))
        goto exit;

    op->operator = operator;

    op->left = left;
    g_object_ref(G_OBJECT(op->left));

    op->right = right;
    g_object_ref(G_OBJECT(op->right));

    result = true;

 exit:

    return result;

}



/* ---------------------------------------------------------------------------------- */
/*                       IMPLEMENTATION DES FONCTIONS DE CLASSE                       */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : item   = premier objet à consulter pour une comparaison.     *
*                other  = second objet à consulter pour une comparaison.      *
*                op     = opération de comparaison à réaliser.                *
*                status = bilan des opérations de comparaison. [OUT]          *
*                                                                             *
*  Description : Réalise une comparaison entre objets selon un critère précis.*
*                                                                             *
*  Retour      : true si la comparaison a pu être effectuée, false sinon.     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_arithmetic_operation_compare_rich(const GArithmOperation *item, const GArithmOperation *other, RichCmpOperation op, bool *status)
{
    bool result;                            /* Etat à retourner            */

    result = g_type_is_a(G_TYPE_FROM_INSTANCE(other), G_TYPE_ARITHMETIC_OPERATION);
    if (!result) goto done;

    if (item->operator != other->operator)
    {
        result = compare_rich_integer_values(item->operator, other->operator, op);
        goto done;
    }

    result = g_comparable_item_compare_rich(G_COMPARABLE_ITEM(item), G_COMPARABLE_ITEM(other), RCO_EQ, status);
    if (!result || STATUS_NOT_EQUAL(*status, op)) goto done;

    result = g_comparable_item_compare_rich(G_COMPARABLE_ITEM(item->left),
                                            G_COMPARABLE_ITEM(other->left),
                                            op, status);
    if (!result || STATUS_NOT_EQUAL(*status, op)) goto done;

    result = g_comparable_item_compare_rich(G_COMPARABLE_ITEM(item->right),
                                            G_COMPARABLE_ITEM(other->right),
                                            op, status);

 done:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : dest = emplacement d'enregistrement à constituer. [OUT]      *
*                src  = expression source à copier.                           *
*                                                                             *
*  Description : Reproduit une expression en place dans une nouvelle instance.*
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arithmetic_operation_copy(GArithmOperation *dest, const GArithmOperation *src)
{
    GScanExpressionClass *class;            /* Classe parente à solliciter */

    class = G_SCAN_EXPRESSION_CLASS(g_arithmetic_operation_parent_class);

    class->copy(G_SCAN_EXPRESSION(dest), G_SCAN_EXPRESSION(src));

    dest->operator = src->operator;

    dest->left = g_scan_expression_duplicate(src->left);
    dest->right = g_scan_expression_duplicate(src->right);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr  = expression à consulter.                              *
*                ctx   = contexte de suivi de l'analyse courante.             *
*                scope = portée courante des variables locales.               *
*                out   = zone d'enregistrement de la réduction opérée. [OUT]  *
*                                                                             *
*  Description : Réduit une expression à une forme plus simple.               *
*                                                                             *
*  Retour      : Bilan de l'opération : false en cas d'erreur irrécupérable.  *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_arithmetic_operation_reduce(GArithmOperation *expr, GScanContext *ctx, GScanScope *scope, GScanExpression **out)
{
    bool result;                            /* Bilan à retourner           */
    GScanExpression *new_left;              /* Expression réduite (gauche) */
    GScanExpression *new_right;             /* Expression réduite (droite) */
    GLiteralExpression *op_left;            /* Opérande gauche final       */
    GLiteralExpression *op_right;           /* Opérande droite final       */
    unsigned long long val_1;               /* Première valeur à traiter   */
    unsigned long long val_2;               /* Seconde valeur à traiter    */
    unsigned long long reduced;             /* Valeur réduite finale       */

    /* Réduction des éléments considérés */

    new_left = NULL;
    new_right = NULL;

    result = g_scan_expression_reduce(expr->left, ctx, scope, &new_left);
    if (!result) goto exit;

    result = g_scan_expression_reduce(expr->right, ctx, scope, &new_right);
    if (!result) goto exit;

    /* Construction d'une réduction locale ? */

    if (G_IS_LITERAL_EXPRESSION(new_left) && G_IS_LITERAL_EXPRESSION(new_right))
    {
        op_left = G_LITERAL_EXPRESSION(new_left);
        op_right = G_LITERAL_EXPRESSION(new_right);

        result = g_literal_expression_get_integer_value(op_left, &val_1);
        if (!result) goto exit;

        result = g_literal_expression_get_integer_value(op_right, &val_2);
        if (!result) goto exit;

        switch (expr->operator)
        {
            case AEO_PLUS:
                reduced = val_1 + val_2;
                break;

            case AEO_MINUS:
                reduced = val_1 - val_2;
                break;

            case AEO_MUL:
                reduced = val_1 * val_2;
                break;

            case AEO_DIV:
                result = (val_2 != 0);
                if (result)
                    reduced = val_1 / val_2;
                break;

            case AEO_MOD:
                result = (val_2 != 0);
                if (result)
                    reduced = val_1 % val_2;
                break;

        }

        if (result)
            *out = g_literal_expression_new(EVT_INTEGER, &reduced);

    }

    /* Mise à jour de la progression ? */

    else if ((new_left != NULL && new_left != expr->left) || (new_right != NULL && new_right != expr->right))
    {
        if (new_left == NULL)
        {
            new_left = expr->left;
            g_object_ref(G_OBJECT(new_left));
        }

        if (new_right == NULL)
        {
            new_right = expr->right;
            g_object_ref(G_OBJECT(new_right));
        }

        *out = g_arithmetic_operation_new(expr->operator, new_left, new_right);

    }

 exit:

    g_clear_object(&new_left);
    g_clear_object(&new_right);

    return result;

}
